// ---------------------------------------------------------------------------
// Define the package we belong to
// ---------------------------------------------------------------------------
package com.mcp.stock.portfolios;

// ---------------------------------------------------------------------------
// Import the Java Packages we are going to need
// ---------------------------------------------------------------------------
import java.io.*;
import java.util.*;
import java.util.EventObject.*;

import com.mcp.stock.portfolios.event.*;


// ---------------------------------------------------------------------------
// public class: PortfolioManager
// ---------------------------------------------------------------------------
public class PortfolioManager
{
    // Create a vector of portfolios
    Vector m_vectorPortfolios = new Vector();
    
    
    // Create our thread that is going to do all of our stock updates
    StockUpdateThread m_threadStockUpdate = new StockUpdateThread( this );
    
    
    // -----------------------------------------------------------------------
    // Constructor
    // -----------------------------------------------------------------------
    public PortfolioManager()
    {
    }
    
    
    // -----------------------------------------------------------------------
    // public Boolean loadAllPortfolios( strFilename )
    //
    //  Loads a vector of Portfolios from their names, which are listed
    //  in strFilename.
    // -----------------------------------------------------------------------
    public Boolean loadAllPortfolios( String strFilename )
    {
        try
        {
            // Create a FileReader to read the portfolio manager file
            FileReader file = new FileReader( strFilename );
            
            // Get the first character from the file
            int nChar = file.read();
            
            // Loop through the file until we get an end of file 
            // marker (-1)
            while( nChar != -1 )
            {
                // String to hold the portfolio names
                String strPortfolio = new String();
                
                // Clear out any carriage returns or linefeeds:
                //  (1) Carriage Return = '\r'
                //  (2) Newline = '\n'
                while( nChar == '\n' || nChar == '\r' )
                {   
                    nChar = file.read();
                }
            
                
                // Read a line from the file
                while( nChar != '\r' && nChar != '\n' && nChar != -1 )
                {
                    // Append this character to our string
                    strPortfolio = strPortfolio + ( char )nChar;
                    
                    // Get the next character
                    nChar = file.read();
                }
                
                // Make sure we read something!
                if( strPortfolio.length() != 0 )
                {
                    // Create a new Portfolio Object
                    Portfolio portfolio = new Portfolio( strPortfolio );
                
                    // Tell the Portfolio to load itself
                    portfolio.loadPortfolio();
                    
                    // Add this portfolio to the Portfolio vector
                    m_vectorPortfolios.addElement( portfolio );
                }
            }
            
        }
        catch( FileNotFoundException e )
        {
            // Not necessarily an error, this just means that there
            // are no portfolios yet -- could be the first time the
            // user ran the application
            return new Boolean( false );
        }
        catch( IOException e )
        {
            // A FileReader.read method call failed!
            System.out.println
            ( 
                "PortfolioManager.loadAllPortfolios Error: " + 
                e.toString() 
            );
            
            return new Boolean( false );
        }

        // Success! Return true
        return new Boolean( true );
    }
    

    // -----------------------------------------------------------------------
    // public Boolean saveAllPortfolios( strFilename )
    //
    //  Save the list of portfolio names to strFilename and tell all
    //  portfolios to save themselves.
    // -----------------------------------------------------------------------
    public Boolean saveAllPortfolios( String strFilename )
    {
        try
        {
            // Create a file that we are going to write to
            FileWriter file = new FileWriter( strFilename );
            
            for( int i=0; i<m_vectorPortfolios.size(); i++ )
            {
                // Get the i-th portfolio from our vector
                Portfolio portfolio = ( Portfolio )m_vectorPortfolios.elementAt( i );
                
                // Tell the portfolio to save itself
                portfolio.savePortfolio();
                
                // Save the portfolio name in our portfolio 
                // manager file
                file.write( portfolio.getName() );
                
                // append a carriage return and line feed to the end
                // of this line
                file.write( "\r\n" );
            }
            
            
            // Close the FileWriter object
            file.close();
        }
        catch( IOException e )
        {
            System.out.println
            (
                "PortfolioManager.saveAllPortfolio Error: " +
                e.toString()
            );
        }
        
        // Success! Return true
        return new Boolean( true );
    }
    
    
    // -----------------------------------------------------------------------
    // public int getNumberOfPortfolios()
    //
    //  Return the number of portfolios in the portfolio manager
    // -----------------------------------------------------------------------
    public int getNumberOfPortfolios()
    {
        return m_vectorPortfolios.size();
    }
    
    
    // -----------------------------------------------------------------------
    // public Portfolio getPortfolio( nIndex )
    //
    //  Retrieves the nIndex-th portfolio from the portfolio manager.
    //  Throws ArrayIndexOutOfBoundsException on error!
    // -----------------------------------------------------------------------
    public Portfolio getPortfolio( int nIndex )
    {
        // See if the index the user is requesting is valid
        if( nIndex >= m_vectorPortfolios.size() )
        {
            // Index is not valid, generate an 
            // ArrayIndexOutOfBoundsException
            ArrayIndexOutOfBoundsException e = 
                new ArrayIndexOutOfBoundsException
                ( 
                    "Requested Index: " + 
                    Integer.toString( nIndex ) +
                    " of a vector with " + 
                    Integer.toString( m_vectorPortfolios.size() ) +
                    " elements"
                );
                
            throw e;
        }
        
        // Return the StockData Object at the given index
        return ( Portfolio )m_vectorPortfolios.elementAt( nIndex );
    }
    
    // -----------------------------------------------------------------------
    // public Portfolio getPortfolio( strName )
    //
    //  Retrieve the portfolio by name.
    //  Throws NoSuchElementException on error
    // -----------------------------------------------------------------------
    public Portfolio getPortfolio( String strName )
    {
        try
        {
            // Loop through the portfolio vector
            for( int i=0; i<m_vectorPortfolios.size(); i++ )
            {
                // Get the Portfolio object form the vector at i
                Portfolio p = ( Portfolio )m_vectorPortfolios.elementAt( i );
                
                // Do we have a match?
                if( p.getName() == strName )
                {
                    // Match! Return this object
                    return p;
                }
            }
        }
        catch( ArrayIndexOutOfBoundsException e )
        {
            // Should never get here!
            System.out.println( "PortfolioManager.getPortfolio Error: " + e.toString() );
        }
        
        // Create a NoSuchElementException - we didn't it in this
        // element in our vector
        NoSuchElementException e = 
            new NoSuchElementException( "Could not find " + strName  + " in portfolio " );
            
        // Throw our exception
        throw e;
    }
    
    
    // -----------------------------------------------------------------------
    // public Boolean updatePortfolioStockPrices()
    //
    //  Tell all of the portfolios in the portfolio manager to update
    //  their stock prices
    // -----------------------------------------------------------------------
    public Boolean updatePortfolioStockPrices()
    {
        // Start our update thread
        m_threadStockUpdate.start();
        
        // Success! Return true
        return new Boolean( true );
    }
    
    // -----------------------------------------------------------------------
    // public void addPortfolio( portfolio )
    //
    //  Add a portfolio to our vector
    // -----------------------------------------------------------------------
    public void addPortfolio( Portfolio portfolio )
    {
        m_vectorPortfolios.addElement( portfolio );
    }
    
    
    // -----------------------------------------------------------------------
    // public void removePortfolio( nIndex )
    //
    //  Remove the portfolio at index nIndex from the vector.
    //  Throws ArrayIndexOutOfBoundsExceptions on error.
    // -----------------------------------------------------------------------
    public void removePorfolio( int nIndex )
    {
        try
        {
            m_vectorPortfolios.removeElementAt( nIndex );
        }
        catch( ArrayIndexOutOfBoundsException e )
        {
            throw e;
        }
    }
    
    // -----------------------------------------------------------------------
    // addStockUpdateListener( StockUpdateListener l )
    //
    //  Add l to the list of listeneres for this object
    // -----------------------------------------------------------------------
    public void addStockUpdateListener( StockUpdateListener l )
    {
        m_threadStockUpdate.addStockUpdateListener( l );
    }
    
    
    // -----------------------------------------------------------------------
    // removeStockUpdateListener( StockUpdateListener l )
    //
    //  Remove l from the list of listeners for this object
    // -----------------------------------------------------------------------
    public void removeStockUpdateListener( StockUpdateListener l )
    {
        m_threadStockUpdate.removeStockUpdateListener( l );
    }
    
    
    // -----------------------------------------------------------------------
    // StockUpdateThread
    //
    //  Used to update all stock prices for all of the portfolios in
    //  the portfolio manager
    // -----------------------------------------------------------------------
    protected class StockUpdateThread extends Thread
    {
        // Reference to our parent
        private PortfolioManager m_portfolioManager;
        
        // Vector of StockUpdateListener Objects
        private Vector m_vStockUpdateListeners = new Vector();
        
        
        // -------------------------------------------------------------------
        // Constructor
        // -------------------------------------------------------------------
        public StockUpdateThread( PortfolioManager portfolioManager )
        {
            m_portfolioManager = portfolioManager;
        }

        
        // -------------------------------------------------------------------
        // addStockUpdateListener( StockUpdateListener l )
        //
        //  Add a StockUpdateListener to this object
        // -------------------------------------------------------------------
        public void addStockUpdateListener( StockUpdateListener l )
        {
            // See if this listener is already in our Vector
            if( m_vStockUpdateListeners.contains( l ) )
            {
                return;
            }
            
            // Add this listener to our Vector
            m_vStockUpdateListeners.add( l );
        }
        
        
        // -------------------------------------------------------------------
        // removeStockUpdateListener( StockUpdateListener l )
        //
        //  Remove a StockUpdateListener from this object
        // -------------------------------------------------------------------
        public void removeStockUpdateListener( StockUpdateListener l )
        {
            // Remove this listener from our Vector
            m_vStockUpdateListeners.removeElement( l );
        }
        
        
        // -------------------------------------------------------------------
        // firePortfolioComplete()
        //
        //  Notify our listeners that a portfolio is complete
        // -------------------------------------------------------------------
        private void firePortfolioComplete()
        {
            // Build a StockUpdateEvent Object
            StockUpdateEvent event = new StockUpdateEvent( m_portfolioManager );
            
            // Notify all of our Listeners
            for( int i=0; i<m_vStockUpdateListeners.size(); i++ )
            {
                StockUpdateListener listener = ( StockUpdateListener )m_vStockUpdateListeners.elementAt( i );
                listener.portfolioComplete( event );
            }
        }
        
        
        // -------------------------------------------------------------------
        // firePortfolioManagerComplete()
        //
        //  Notify our listeners that the portfolio manager is complete
        // -------------------------------------------------------------------
        private void firePortfolioManagerComplete()
        {
            // Build a StockUpdateEvent Object
            StockUpdateEvent event = new StockUpdateEvent( m_portfolioManager );
            
            // Notify all of our Listeners
            for( int i=0; i<m_vStockUpdateListeners.size(); i++ )
            {
                StockUpdateListener listener = ( StockUpdateListener )m_vStockUpdateListeners.elementAt( i );
                listener.portfolioManagerComplete( event );
            }
        }
        
        
        // -------------------------------------------------------------------
        // public void run()
        //
        //  Main body of the thread that is called when this thread 
        //  is started
        // -------------------------------------------------------------------
        public void run()
        {
            // Loop through all of the portfolios in the vector
            for( int i=0; i<m_vectorPortfolios.size(); i++ )
            {
                // Retrieve the i-th portfolio
                Portfolio portfolio = ( Portfolio )m_vectorPortfolios.elementAt( i );
                
                // Tell it to update its stock price
                if( portfolio.updateStockPrices().booleanValue() == false )
                {
                    // An error occurred!
                    return;
                }
                
                // Notify our listeners that a portfolio has completed
                firePortfolioComplete();
            }

            // Notify our action listener that the stock update is complete
            firePortfolioManagerComplete();
        }
        
    }
    
}

